class ManifestooException(Exception):
    """Base class of all manifestoo_core exceptions."""


class UnsupportedOdooVersion(ManifestooException):
    pass


class UnsupportedManifestVersion(ManifestooException):
    pass


class InvalidManifest(ManifestooException):
    pass


class AddonNotFound(ManifestooException):
    pass


class AddonNotFoundNotInstallable(AddonNotFound):
    pass


class AddonNotFoundNoInit(AddonNotFound):
    pass


class AddonNotFoundNoManifest(AddonNotFound):
    pass


class AddonNotFoundNotADirectory(AddonNotFound):
    pass


class AddonNotFoundInvalidManifest(AddonNotFound):
    pass



import ast
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, TypeVar

from .exceptions import InvalidManifest

T = TypeVar("T")
VT = TypeVar("VT")

__all__ = ["Manifest", "InvalidManifest", "get_manifest_path"]


def _check_str(value: Any) -> str:
    if not isinstance(value, str):
        raise TypeError()
    return value


def _check_optional_str(value: Any) -> Optional[str]:
    if value is None:
        return value
    return _check_str(value)


def _check_bool(value: Any) -> bool:
    if not isinstance(value, bool):
        raise TypeError()
    return value


def _check_list(value: Any, item_checker: Callable[[Any], T]) -> List[T]:
    if not isinstance(value, list):
        raise TypeError()
    map(item_checker, value)
    return value


def _check_dict(
    value: Any, key_checker: Callable[[Any], T], value_checker: Callable[[Any], VT]
) -> Dict[T, VT]:
    if not isinstance(value, dict):
        raise TypeError()
    for k, v in value.items():
        key_checker(k)
        value_checker(v)
    return value


def _check_list_of_str(value: Any) -> List[str]:
    # this could be a partial but mypy does not support it
    return _check_list(value, item_checker=_check_str)


def _check_dict_of_list_of_str(value: Any) -> Dict[str, List[str]]:
    # this could be a partial but mypy does not support it
    return _check_dict(value, key_checker=_check_str, value_checker=_check_list_of_str)


def get_manifest_path(addon_dir: Path) -> Optional[Path]:
    """Get the path to the manifest file for an addon directory.

    Returns None if no manifest file is found.
    """
    for manifest_name in ("__manifest__.py", "__openerp__.py", "__terp__.py"):
        manifest_path = addon_dir / manifest_name
        if manifest_path.is_file():
            return manifest_path
    return None


class Manifest:
    def __init__(self, manifest_dict: Dict[str, Any]) -> None:
        self.manifest_dict = manifest_dict

    def _get(self, key: str, checker: Callable[[Any], T], default: T) -> T:
        """Get value with runtime type check."""
        try:
            value = self.manifest_dict[key]
        except KeyError:
            return default
        try:
            return checker(value)
        except TypeError:
            raise InvalidManifest(f"{value!r} has invalid type for {key!r} in {self}")

    @property
    def name(self) -> Optional[str]:
        return self._get("name", _check_optional_str, default=None)

    @property
    def version(self) -> Optional[str]:
        return self._get("version", _check_optional_str, default=None)

    @property
    def installable(self) -> bool:
        return self._get("installable", _check_bool, default=True)

    @property
    def depends(self) -> List[str]:
        return self._get("depends", _check_list_of_str, default=[])

    @property
    def external_dependencies(self) -> Dict[str, List[str]]:
        return self._get(
            "external_dependencies", _check_dict_of_list_of_str, default={}
        )

    @property
    def license(self) -> Optional[str]:
        return self._get("license", _check_optional_str, default=None)

    @property
    def development_status(self) -> Optional[str]:
        return self._get("development_status", _check_optional_str, default=None)

    @classmethod
    def from_dict(
        cls, manifest_dict: Dict[Any, Any], source: str = "<manifest>"
    ) -> "Manifest":
        """Parse a manifest dictionary into a :class:`Manifest` object.

        Raises :class:`InvalidManifest` if the manifest is invalid.
        """
        if not isinstance(manifest_dict, dict):
            raise InvalidManifest(f"Manifest {source} is not a dictionary")
        if any(not isinstance(k, str) for k in manifest_dict):
            raise InvalidManifest(f"Manifest {source} has non-string keys")
        return cls(manifest_dict)

    @classmethod
    def from_str(cls, manifest_str: str, source: str = "<manifest>") -> "Manifest":
        """Parse a manifest string into a :class:`Manifest` object.

        Raises :class:`InvalidManifest` if the manifest is invalid.
        """
        try:
            manifest_dict = ast.literal_eval(manifest_str)
        except SyntaxError as e:
            raise InvalidManifest(f"Manifest {source} has invalid syntax") from e
        return cls.from_dict(manifest_dict)

    @classmethod
    def from_file(cls, manifest_path: Path) -> "Manifest":
        """Parse a manifest file into a :class:`Manifest` object.

        Raises :class:`InvalidManifest` if the manifest is invalid.
        """
        manifest_str = manifest_path.read_text()
        return cls.from_str(manifest_str, source=f"{manifest_path!r}")



import subprocess
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pytest
from pkg_metadata import msg_to_json

from manifestoo_core.exceptions import (
    UnsupportedManifestVersion,
    UnsupportedOdooVersion,
)
from manifestoo_core.metadata import (
    POST_VERSION_STRATEGY_DOT_N,
    POST_VERSION_STRATEGY_NINETYNINE_DEVN,
    POST_VERSION_STRATEGY_NONE,
    POST_VERSION_STRATEGY_P1_DEVN,
    metadata_from_addon_dir,
)


def _no_none(d: Dict[str, Any]) -> Dict[str, Any]:
    return {k: v for k, v in d.items() if v is not None}


def _m(
    tmp_path: Path,
    *,
    addon_dir_name: str = "addon1",
    # manifest
    name: Optional[str] = "Addon 1",
    version: str = "14.0.1.0.0",
    summary: Optional[str] = None,
    description: Optional[str] = None,
    readme_rst: Optional[str] = None,
    depends: Optional[List[str]] = None,
    external_dependencies: Optional[Dict[str, List[str]]] = None,
    website: Optional[str] = None,
    author: Optional[str] = None,
    license: Optional[str] = None,
    development_status: Optional[str] = None,
    # options
    depends_override: Optional[Dict[str, str]] = None,
    external_dependencies_override: Optional[
        Dict[str, Dict[str, Union[str, List[str]]]]
    ] = None,
    odoo_version_override: Optional[str] = None,
    post_version_strategy_override: Optional[str] = None,
    precomputed_metadata_file: Optional[Path] = None,
) -> Dict[str, Any]:
    addon_dir = tmp_path / addon_dir_name
    addon_dir.mkdir()
    addon_dir.joinpath("__init__.py").touch()
    manifest_path = addon_dir / "__manifest__.py"
    manifest_path.write_text(
        repr(
            _no_none(
                dict(
                    name=name,
                    version=version,
                    summary=summary,
                    description=description,
                    depends=depends,
                    external_dependencies=external_dependencies,
                    website=website,
                    author=author,
                    license=license,
                    development_status=development_status,
                )
            )
        )
    )
    if readme_rst:
        readme_path = addon_dir / "README.rst"
        readme_path.write_text(readme_rst)
    return msg_to_json(
        metadata_from_addon_dir(
            addon_dir,
            options=dict(
                depends_override=depends_override,
                external_dependencies_override=external_dependencies_override,
                odoo_version_override=odoo_version_override,
                post_version_strategy_override=post_version_strategy_override,
            ),
            precomputed_metadata_file=precomputed_metadata_file,
        )
    )


def test_basic(tmp_path: Path) -> None:
    assert _m(tmp_path) == dict(
        name="odoo14-addon-addon1",
        version="14.0.1.0.0",
        summary="Addon 1",
        requires_dist=["odoo>=14.0a,<14.1dev"],
        requires_python=">=3.6",
        classifier=[
            "Programming Language :: Python",
            "Framework :: Odoo",
            "Framework :: Odoo :: 14.0",
        ],
        metadata_version="2.2",
    )


@pytest.mark.parametrize(
    ("odoo_series", "expected"),
    [
        ("8.0", ["odoo>=8.0a,<9.0a"]),
        ("9.0", ["odoo>=9.0a,<9.1a"]),
        ("10.0", ["odoo>=10.0,<10.1dev"]),
        ("11.0", ["odoo>=11.0a,<11.1dev"]),
        ("12.0", ["odoo>=12.0a,<12.1dev"]),
        ("13.0", ["odoo>=13.0a,<13.1dev"]),
        ("14.0", ["odoo>=14.0a,<14.1dev"]),
        ("15.0", ["odoo>=15.0a,<15.1dev"]),
        ("16.0", ["odoo>=16.0a,<16.1dev"]),  # TODO >=16.0dev ?
    ],
)
def test_requires_odoo(tmp_path: Path, odoo_series: str, expected: List[str]) -> None:
    assert _m(tmp_path, version=f"{odoo_series}.1.0.0")["requires_dist"] == expected


@pytest.mark.parametrize(
    ("odoo_series", "expected"),
    [
        ("8.0", "~=2.7"),
        ("9.0", "~=2.7"),
        ("10.0", "~=2.7"),
        ("11.0", ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*"),
        ("12.0", ">=3.5"),
        ("13.0", ">=3.5"),
        ("14.0", ">=3.6"),
        ("15.0", ">=3.8"),
        ("16.0", ">=3.10"),
    ],
)
def test_requires_python(tmp_path: Path, odoo_series: str, expected: str) -> None:
    assert _m(tmp_path, version=f"{odoo_series}.1.0.0")["requires_python"] == expected


@pytest.mark.parametrize(
    ("odoo_series", "expected_prefix"),
    [
        ("8.0", "odoo8-addon-"),
        ("9.0", "odoo9-addon-"),
        ("10.0", "odoo10-addon-"),
        ("11.0", "odoo11-addon-"),
        ("12.0", "odoo12-addon-"),
        ("13.0", "odoo13-addon-"),
        ("14.0", "odoo14-addon-"),
        ("15.0", "odoo-addon-"),
        ("16.0", "odoo-addon-"),
    ],
)
def test_name_prefix(tmp_path: Path, odoo_series: str, expected_prefix: str) -> None:
    assert (
        _m(tmp_path, name="addon1", version=f"{odoo_series}.1.0.0")["name"]
        == f"{expected_prefix}addon1"
    )


def test_depends_core_addon(tmp_path: Path) -> None:
    """A dependency on a core addon should be ignored."""
    assert _m(tmp_path, version="14.0.1.0.0", depends=["base"])["requires_dist"] == [
        "odoo>=14.0a,<14.1dev"
    ]


@pytest.mark.parametrize(
    ("odoo_series", "depends", "expected"),
    [
        ("8.0", ["mis_builder"], ["odoo8-addon-mis_builder"]),
        ("9.0", ["mis_builder"], ["odoo9-addon-mis_builder"]),
        ("10.0", ["mis_builder"], ["odoo10-addon-mis_builder"]),
        ("11.0", ["mis_builder"], ["odoo11-addon-mis_builder"]),
        ("12.0", ["mis_builder"], ["odoo12-addon-mis_builder"]),
        ("13.0", ["mis_builder"], ["odoo13-addon-mis_builder"]),
        ("14.0", ["mis_builder"], ["odoo14-addon-mis_builder"]),
        ("15.0", ["mis_builder"], ["odoo-addon-mis_builder>=15.0dev,<15.1dev"]),
        ("16.0", ["mis_builder"], ["odoo-addon-mis_builder>=16.0dev,<16.1dev"]),
        (
            "14.0",
            [
                "mis_builder",
                "mis_builder_budget",
            ],
            [
                "odoo14-addon-mis_builder",
                "odoo14-addon-mis_builder_budget",
            ],
        ),
        (
            "16.0",
            [
                "base",
                "mis_builder",
                "mis_builder_budget",
            ],
            [
                "odoo-addon-mis_builder>=16.0dev,<16.1dev",
                "odoo-addon-mis_builder_budget>=16.0dev,<16.1dev",
            ],
        ),
    ],
)
def test_depends_noncore_addon(
    tmp_path: Path, odoo_series: str, depends: List[str], expected: List[str]
) -> None:
    """A dependency on a non-core addon appears in requires_dist."""
    requires_dist = _m(tmp_path, version=f"{odoo_series}.1.0.0", depends=depends)[
        "requires_dist"
    ]
    assert [d for d in requires_dist if not d.startswith("odoo>=")] == expected


def test_depends_override(tmp_path: Path) -> None:
    """A dependency on a non-core addon appears in requires_dist."""
    assert _m(
        tmp_path,
        depends=["mis_builder"],
        depends_override={"mis_builder": "odoo14-addon-mis_builder>=14.0.4.0.0"},
    )["requires_dist"] == [
        "odoo14-addon-mis_builder>=14.0.4.0.0",
        "odoo>=14.0a,<14.1dev",
    ]


def test_external_dependencies(tmp_path: Path) -> None:
    assert _m(tmp_path, external_dependencies={"python": ["lxml"]})[
        "requires_dist"
    ] == [
        "lxml",
        "odoo>=14.0a,<14.1dev",
    ]


def test_external_dependencies_override(tmp_path: Path) -> None:
    assert _m(
        tmp_path,
        external_dependencies={"python": ["lxml"]},
        external_dependencies_override={"python": {"lxml": "lxml>=3.8.0"}},
    )["requires_dist"] == [
        "lxml>=3.8.0",
        "odoo>=14.0a,<14.1dev",
    ]


def test_external_dependencies_override_multi(tmp_path: Path) -> None:
    assert _m(
        tmp_path,
        external_dependencies={"python": ["lxml"]},
        external_dependencies_override={
            "python": {"lxml": ["lxml>=3.8.0", "something"]}
        },
    )["requires_dist"] == ["lxml>=3.8.0", "odoo>=14.0a,<14.1dev", "something"]


def test_odoo_version_unsupported(tmp_path: Path) -> None:
    with pytest.raises(UnsupportedOdooVersion):
        _m(tmp_path, version="45.0.1.0.0")


@pytest.mark.parametrize("version", ["1", "1.0", "1.0.0", "1.0.0.0", "10.0.0.0"])
def test_manifest_version_undetermined(tmp_path: Path, version: str) -> None:
    with pytest.raises(UnsupportedManifestVersion):
        _m(tmp_path, version=version)


def test_odoo_version_override(tmp_path: Path) -> None:
    assert _m(tmp_path, version="45.0.1.0.0", odoo_version_override="14.0")[
        "requires_dist"
    ] == ["odoo>=14.0a,<14.1dev"]


def test_summary_defaults_to_name(tmp_path: Path) -> None:
    assert _m(tmp_path, name="addon1")["summary"] == "addon1"


def test_summary_from_summary(tmp_path: Path) -> None:
    assert _m(tmp_path, name="addon1", summary="Addon 1")["summary"] == "Addon 1"


def test_description(
    tmp_path: Path, description: str = "A description\n\nwith two lines"
) -> None:
    assert _m(tmp_path, description=description)["description"] == description


def test_description_from_readme(
    tmp_path: Path, readme_rst: str = "A readme\n\nwith two lines"
) -> None:
    assert _m(tmp_path, readme_rst=readme_rst)["description"] == readme_rst


def test_description_from_description_and_readme(
    tmp_path: Path,
    description: str = "A description",
    readme_rst: str = "A readme\n\nwith two lines",
) -> None:
    assert (
        _m(tmp_path, description=description, readme_rst=readme_rst)["description"]
        == readme_rst
    )


def test_author(tmp_path: Path) -> None:
    assert _m(tmp_path, author="John Doe")["author"] == "John Doe"


def test_author_no_nl(tmp_path: Path) -> None:
    assert _m(tmp_path, author="John Doe,\nOCA")["author"] == "John Doe, OCA"


def test_author_email_oca(tmp_path: Path) -> None:
    assert "author_email" not in _m(
        tmp_path,
        addon_dir_name="a",
        author="John Doe",
    )
    assert (
        _m(
            tmp_path,
            addon_dir_name="b",
            author="John Doe, Odoo Community Association (OCA)",
        )["author_email"]
        == "support@odoo-community.org"
    )


@pytest.mark.parametrize(
    "odoo_series",
    ["8.0", "9.0", "10.0", "11.0", "12.0", "13.0", "14.0", "15.0", "16.0"],
)
def test_classifiers(tmp_path: Path, odoo_series: str) -> None:
    assert _m(tmp_path, version=f"{odoo_series}.1.0.0")["classifier"] == [
        "Programming Language :: Python",
        "Framework :: Odoo",
        f"Framework :: Odoo :: {odoo_series}",
    ]


@pytest.mark.parametrize(
    ("license", "expected_license"),
    [
        (
            "agpl-3",
            "License :: OSI Approved :: GNU Affero General Public License v3",
        ),
        (
            "AGPL-3",
            "License :: OSI Approved :: GNU Affero General Public License v3",
        ),
        (
            "agpl-3 or any later version",
            "License :: OSI Approved :: "
            "GNU Affero General Public License v3 or later (AGPLv3+)",
        ),
        (
            "gpl-2",
            "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        ),
        (
            "gpl-2 or any later version",
            "License :: OSI Approved :: "
            "GNU General Public License v2 or later (GPLv2+)",
        ),
        (
            "gpl-3",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        ),
        (
            "gpl-3 or any later version",
            "License :: OSI Approved :: "
            "GNU General Public License v3 or later (GPLv3+)",
        ),
        (
            "lgpl-2",
            "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        ),
        (
            "lgpl-2 or any later version",
            "License :: OSI Approved :: "
            "GNU Lesser General Public License v2 or later (LGPLv2+)",
        ),
        (
            "lgpl-3",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        ),
        (
            "lgpl-3 or any later version",
            "License :: OSI Approved :: "
            "GNU Lesser General Public License v3 or later (LGPLv3+)",
        ),
    ],
)
def test_classifiers_license(
    tmp_path: Path, license: str, expected_license: str
) -> None:
    assert expected_license in _m(tmp_path, license=license)["classifier"]


@pytest.mark.parametrize(
    ("development_status", "expected_development_status"),
    [
        ("alpha", "Development Status :: 3 - Alpha"),
        ("beta", "Development Status :: 4 - Beta"),
        ("production/stable", "Development Status :: 5 - Production/Stable"),
        ("stable", "Development Status :: 5 - Production/Stable"),
        ("production", "Development Status :: 5 - Production/Stable"),
        ("mature", "Development Status :: 6 - Mature"),
    ],
)
def test_classifiers_development_status(
    tmp_path: Path, development_status: str, expected_development_status: str
) -> None:
    assert (
        expected_development_status
        in _m(tmp_path, development_status=development_status)["classifier"]
    )


def test_license(tmp_path: Path, license: str = "AGPL-3") -> None:
    assert _m(tmp_path, license=license)["license"] == license


def test_home_page(tmp_path: Path, website: str = "https://acsone.eu") -> None:
    assert _m(tmp_path, website=website)["home_page"] == website


def test_precomputed_metadata_path(tmp_path: Path) -> None:
    pkg_info_path = tmp_path / "PKG-INFO"
    pkg_info_path.write_text("Name: odoo14-addon-addon1\nVersion: 14.0.1.0.0.3")
    metadata = _m(
        tmp_path,
        addon_dir_name="tmp",
        version="14.0.1.0.0",
        precomputed_metadata_file=pkg_info_path,
    )
    assert metadata["name"] == "odoo14-addon-addon1"
    assert metadata["version"] == "14.0.1.0.0.3"


def _make_git_addon(
    tmp_path: Path,
    manifest_version: str,
    post_commits: int = 0,
    addon_name: str = "addon1",
) -> Path:
    addon_dir = tmp_path / addon_name
    addon_dir.mkdir()
    addon_dir.joinpath("__manifest__.py").write_text(
        f"{{'name': '{addon_name}', 'version': '{manifest_version}'}}"
    )
    addon_dir.joinpath("__init__.py").touch()
    subprocess.check_call(["git", "init"], cwd=addon_dir)
    subprocess.check_call(
        ["git", "config", "user.email", "test@example.com"], cwd=addon_dir
    )
    subprocess.check_call(["git", "config", "user.name", "test"], cwd=addon_dir)
    subprocess.check_call(["git", "add", "."], cwd=addon_dir)
    subprocess.check_call(["git", "commit", "-m", "initial commit"], cwd=addon_dir)
    for i in range(post_commits):
        addon_dir.joinpath("README.rst").write_text(f"{i}")
        subprocess.check_call(["git", "add", "README.rst"], cwd=addon_dir)
        subprocess.check_call(["git", "commit", "-m", f"commit {i}"], cwd=addon_dir)
    return addon_dir


@pytest.mark.parametrize(
    (
        "manifest_version",
        "post_commits",
        "post_version_strategy_override",
        "expected_version",
    ),
    [
        # last commit is manifest version change
        ("8.0.1.0.0", 0, None, "8.0.1.0.0"),
        ("9.0.1.0.0", 0, None, "9.0.1.0.0"),
        ("10.0.1.0.0", 0, None, "10.0.1.0.0"),
        ("11.0.1.0.0", 0, None, "11.0.1.0.0"),
        ("12.0.1.0.0", 0, None, "12.0.1.0.0"),
        ("13.0.1.0.0", 0, None, "13.0.1.0.0"),
        ("14.0.1.0.0", 0, None, "14.0.1.0.0"),
        ("15.0.1.0.0", 0, None, "15.0.1.0.0"),
        ("16.0.1.0.0", 0, None, "16.0.1.0.0"),
        # 2 commits after manifest version change
        ("8.0.1.0.0", 2, None, "8.0.1.0.0.99.dev2"),
        ("9.0.1.0.0", 2, None, "9.0.1.0.0.99.dev2"),
        ("10.0.1.0.0", 2, None, "10.0.1.0.0.99.dev2"),
        ("11.0.1.0.0", 2, None, "11.0.1.0.0.99.dev2"),
        ("12.0.1.0.0", 2, None, "12.0.1.0.0.99.dev2"),
        ("13.0.1.0.0", 2, None, "13.0.1.0.1.dev2"),
        ("14.0.1.0.0", 2, None, "14.0.1.0.1.dev2"),
        ("15.0.1.0.0", 2, None, "15.0.1.0.0.2"),
        ("16.0.1.0.0", 2, None, "16.0.1.0.0.2"),
        # strategy overrides
        ("16.0.1.0.0", 2, POST_VERSION_STRATEGY_NONE, "16.0.1.0.0"),
        ("16.0.1.0.0", 2, POST_VERSION_STRATEGY_NINETYNINE_DEVN, "16.0.1.0.0.99.dev2"),
        ("16.0.1.0.0", 2, POST_VERSION_STRATEGY_P1_DEVN, "16.0.1.0.1.dev2"),
        ("16.0.1.0.0", 2, POST_VERSION_STRATEGY_DOT_N, "16.0.1.0.0.2"),
    ],
)
def test_git_post_version(
    tmp_path: Path,
    manifest_version: str,
    post_commits: int,
    post_version_strategy_override: Optional[str],
    expected_version: str,
) -> None:
    addon_dir = _make_git_addon(
        tmp_path,
        manifest_version=manifest_version,
        post_commits=post_commits,
    )
    metadata = msg_to_json(
        metadata_from_addon_dir(
            addon_dir,
            options=dict(post_version_strategy_override=post_version_strategy_override),
        )
    )
    assert metadata["version"] == expected_version


@pytest.mark.parametrize(
    (
        "manifest_version",
        "post_commits",
        "post_version_strategy_override",
        "expected_version",
    ),
    [
        ("16.0.1.0.0", 1, POST_VERSION_STRATEGY_NONE, "16.0.1.0.0"),
        ("16.0.1.0.0", 1, POST_VERSION_STRATEGY_NINETYNINE_DEVN, "16.0.1.0.0.99.dev2"),
        ("16.0.1.0.0", 1, POST_VERSION_STRATEGY_P1_DEVN, "16.0.1.0.1.dev2"),
        ("16.0.1.0.0", 1, POST_VERSION_STRATEGY_DOT_N, "16.0.1.0.0.2"),
    ],
)
def test_git_post_version_uncommitted_change(
    tmp_path: Path,
    manifest_version: str,
    post_commits: int,
    post_version_strategy_override: str,
    expected_version: str,
) -> None:
    addon_dir = _make_git_addon(
        tmp_path, manifest_version=manifest_version, post_commits=post_commits
    )
    addon_dir.joinpath("README.rst").write_text("stuff")
    metadata = msg_to_json(
        metadata_from_addon_dir(
            addon_dir,
            options=dict(post_version_strategy_override=post_version_strategy_override),
        )
    )
    assert metadata["version"] == expected_version


@pytest.mark.parametrize(
    "post_version_strategy_override",
    [
        POST_VERSION_STRATEGY_NONE,
        POST_VERSION_STRATEGY_NINETYNINE_DEVN,
        POST_VERSION_STRATEGY_P1_DEVN,
        POST_VERSION_STRATEGY_DOT_N,
    ],
)
def test_git_post_version_bad_manifest_in_history(
    tmp_path: Path,
    post_version_strategy_override: str,
) -> None:
    addon_dir = _make_git_addon(tmp_path, manifest_version="16.0.1.1.0")
    addon_dir.joinpath("__manifest__.py").write_text(
        "{syntaxerror, 'version': '16.0.1.2.0'}"
    )
    subprocess.check_call(["git", "add", "__manifest__.py"], cwd=addon_dir)
    subprocess.check_call(["git", "commit", "-m", "bad manifest"], cwd=addon_dir)
    addon_dir.joinpath("__manifest__.py").write_text(
        "{'name': 'A', 'version': '16.0.1.3.0'}"
    )
    subprocess.check_call(["git", "add", "__manifest__.py"], cwd=addon_dir)
    subprocess.check_call(["git", "commit", "-m", "good manifest"], cwd=addon_dir)
    metadata = msg_to_json(
        metadata_from_addon_dir(
            addon_dir,
            options=dict(post_version_strategy_override=post_version_strategy_override),
        )
    )
    assert metadata["version"] == "16.0.1.3.0"


@pytest.mark.parametrize(
    "post_version_strategy_override",
    [
        POST_VERSION_STRATEGY_NONE,
        POST_VERSION_STRATEGY_NINETYNINE_DEVN,
        POST_VERSION_STRATEGY_P1_DEVN,
        POST_VERSION_STRATEGY_DOT_N,
    ],
)
def test_git_post_version_good_manifest_in_history(
    tmp_path: Path,
    post_version_strategy_override: str,
) -> None:
    addon_dir = _make_git_addon(tmp_path, manifest_version="16.0.1.1.0")
    addon_dir.joinpath("__manifest__.py").write_text(
        "{'name': 'A', 'version': '16.0.1.2.0'}"
    )
    subprocess.check_call(["git", "add", "__manifest__.py"], cwd=addon_dir)
    subprocess.check_call(["git", "commit", "-m", "good manifest"], cwd=addon_dir)
    metadata = msg_to_json(
        metadata_from_addon_dir(
            addon_dir,
            options=dict(post_version_strategy_override=post_version_strategy_override),
        )
    )
    assert metadata["version"] == "16.0.1.2.0"


@pytest.mark.parametrize(
    "post_version_strategy_override",
    [
        POST_VERSION_STRATEGY_NONE,
        POST_VERSION_STRATEGY_NINETYNINE_DEVN,
        POST_VERSION_STRATEGY_P1_DEVN,
        POST_VERSION_STRATEGY_DOT_N,
    ],
)
def test_git_post_version_no_manifest_in_history(
    tmp_path: Path,
    post_version_strategy_override: str,
) -> None:
    addon_dir = _make_git_addon(tmp_path, manifest_version="16.0.1.1.0")
    subprocess.check_call(["git", "rm", "__manifest__.py"], cwd=addon_dir)
    subprocess.check_call(["git", "commit", "-m", "no manifest"], cwd=addon_dir)
    addon_dir.joinpath("__manifest__.py").write_text(
        "{'name': 'A', 'version': '16.0.1.2.0'}"
    )
    subprocess.check_call(["git", "add", "__manifest__.py"], cwd=addon_dir)
    subprocess.check_call(["git", "commit", "-m", "good manifest"], cwd=addon_dir)
    metadata = msg_to_json(
        metadata_from_addon_dir(
            addon_dir,
            options=dict(post_version_strategy_override=post_version_strategy_override),
        )
    )
    assert metadata["version"] == "16.0.1.2.0"



from pathlib import Path
from typing import Optional

from .exceptions import (
    AddonNotFound,
    AddonNotFoundInvalidManifest,
    AddonNotFoundNoInit,
    AddonNotFoundNoManifest,
    AddonNotFoundNotADirectory,
    AddonNotFoundNotInstallable,
)
from .manifest import InvalidManifest, Manifest, get_manifest_path

__all__ = [
    "is_addon_dir",
]


def is_addon_dir(addon_dir: Path, allow_not_installable: bool = False) -> bool:
    """Detect if a directory contains an Odoo addon.

    :param addon_dir: The directory to check.
    :param allow_not_installable: Whether to allow the addon to be have
        installable=False in its manifest.
    """
    try:
        Addon.from_addon_dir(addon_dir, allow_not_installable)
    except AddonNotFound:
        return False
    else:
        return True


class Addon:
    """Represent a concrete addon manifest."""

    def __init__(
        self, manifest: Manifest, manifest_path: Path, name: Optional[str] = None
    ):
        self.manifest = manifest
        self.manifest_path = manifest_path
        self.path = self.manifest_path.parent
        if name is None:
            self.name = self.path.name
        else:
            self.name = name

    @classmethod
    def from_addon_dir(
        cls, addon_dir: Path, allow_not_installable: bool = False
    ) -> "Addon":
        if not addon_dir.is_dir():
            raise AddonNotFoundNotADirectory(f"{addon_dir} is not a directory")
        manifest_path = get_manifest_path(addon_dir)
        if not manifest_path:
            raise AddonNotFoundNoManifest(f"No manifest file found in {addon_dir}")
        try:
            manifest = Manifest.from_file(manifest_path)
            if not allow_not_installable and not manifest.installable:
                raise AddonNotFoundNotInstallable(f"{addon_dir} is not installable")
        except InvalidManifest as e:
            raise AddonNotFoundInvalidManifest(str(e)) from e
        if not addon_dir.joinpath("__init__.py").is_file():
            raise AddonNotFoundNoInit(f"{addon_dir} is missing an __init__.py")
        return cls(manifest, manifest_path)



import sys
from email.message import Message
from pathlib import Path
from typing import Dict, List, Optional, Union

if sys.version_info >= (3, 8):
    from typing import Final, TypedDict
else:
    from typing_extensions import Final, TypedDict

from .addon import Addon
from .exceptions import UnsupportedManifestVersion, UnsupportedOdooVersion

POST_VERSION_STRATEGY_NONE: Final = "none"
POST_VERSION_STRATEGY_NINETYNINE_DEVN: Final = ".99.devN"
POST_VERSION_STRATEGY_P1_DEVN: Final = "+1.devN"
POST_VERSION_STRATEGY_DOT_N: Final = ".N"


class MetadataOptions(TypedDict, total=False):
    """A dictionary of options for metadata conversion.

    Supported keys are:

    - ``depends_override``
    - ``external_dependencies_override``
    - ``odoo_version_override``
    - ``post_version_strategy_override``
    """

    depends_override: Optional[Dict[str, str]]
    external_dependencies_override: Optional[
        Dict[str, Dict[str, Union[str, List[str]]]]
    ]
    odoo_version_override: Optional[str]
    post_version_strategy_override: Optional[str]


def metadata_from_addon_dir(
    addon_dir: Path,
    options: Optional[MetadataOptions] = None,
    precomputed_metadata_file: Optional[Path] = None,
) -> Message:
    """Return Python Package Metadata 2.2 for an Odoo addon directory as an
    ``email.message.Message``.

    The Description field is absent and is stored in the message payload. All values are
    guaranteed to not contain newline characters, except for the payload.

    ``precomputed_metadata_path`` may point to a file containing pre-computed metadata
    that will be used to obtain the Name and Version, instead of looking at the addon
    directory name or manifest version + VCS, respectively. This is useful to process a
    manifest from a sdist tarball with PKG-INFO, for example, when the original
    directory name or VCS is not available to compute the package name and version.

    This function may raise :class:`manifestoo_core.exceptions.ManifestooException` if
    ``addon_dir`` does not contain a valid installable Odoo addon for a supported Odoo
    version.
    """
    Addon.from_addon_dir(addon_dir)  # check we have a proper addon directory
    return _metadata_from_addon_dir_using_setuptools_odoo(
        addon_dir,
        options or MetadataOptions(),
        precomputed_metadata_file,
    )


def _metadata_from_addon_dir_using_setuptools_odoo(
    addon_dir: Path,
    options: MetadataOptions,
    precomputed_metadata_file: Optional[Path] = None,
) -> Message:
    from distutils.errors import DistutilsSetupError

    from setuptools_odoo import get_addon_metadata  # type: ignore

    try:
        return get_addon_metadata(  # type: ignore
            str(addon_dir),
            depends_override=options.get("depends_override"),
            external_dependencies_override=options.get(
                "external_dependencies_override"
            ),
            odoo_version_override=options.get("odoo_version_override"),
            post_version_strategy_override=options.get(
                "post_version_strategy_override"
            ),
            precomputed_metadata_path=precomputed_metadata_file,
        )
    except DistutilsSetupError as e:
        if "Unsupported odoo version" in str(e):
            raise UnsupportedOdooVersion(str(e)) from e
        elif "Version in manifest must" in str(e):
            raise UnsupportedManifestVersion(str(e)) from e
        raise



from pathlib import Path

import pytest
from pytest import FixtureRequest

from manifestoo_core.addon import Addon, is_addon_dir
from manifestoo_core.exceptions import (
    AddonNotFoundInvalidManifest,
    AddonNotFoundNoInit,
    AddonNotFoundNoManifest,
    AddonNotFoundNotADirectory,
    AddonNotFoundNotInstallable,
)
from manifestoo_core.manifest import Manifest


@pytest.fixture(
    params=[
        {
            "dir": "a",
            "manifest": "{}",
        }
    ]
)
def addon_dir(request: FixtureRequest, tmp_path: Path) -> Path:
    addon_dir = tmp_path.joinpath(request.param["dir"])
    addon_dir.mkdir()
    addon_dir.joinpath("__init__.py").touch()
    addon_dir.joinpath("__manifest__.py").write_text(request.param["manifest"])
    return addon_dir


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "theaddon",
            "manifest": "{'name': 'the addon'}",
        }
    ],
    indirect=True,
)
def test_basic(addon_dir: Path) -> None:
    addon = Addon.from_addon_dir(addon_dir)
    assert addon.name == "theaddon"
    assert addon.path == addon_dir
    assert addon.manifest.name == "the addon"


def test_not_a_directory(tmp_path: Path) -> None:
    with pytest.raises(AddonNotFoundNotADirectory):
        Addon.from_addon_dir(tmp_path / "not-a-dir")


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "a",
            "manifest": "{'installable': False}",
        }
    ],
    indirect=True,
)
def test_not_installable(addon_dir: Path) -> None:
    with pytest.raises(AddonNotFoundNotInstallable):
        Addon.from_addon_dir(addon_dir)
    assert Addon.from_addon_dir(addon_dir, allow_not_installable=True)


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "a",
            "manifest": "[]",
        }
    ],
    indirect=True,
)
def test_invalid_manifest(addon_dir: Path) -> None:
    with pytest.raises(AddonNotFoundInvalidManifest):
        Addon.from_addon_dir(addon_dir)


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "a",
            "manifest": "{'installable':}",
        }
    ],
    indirect=True,
)
def test_manifest_syntax_error(addon_dir: Path) -> None:
    with pytest.raises(AddonNotFoundInvalidManifest):
        Addon.from_addon_dir(addon_dir)


@pytest.mark.parametrize(
    "addon_dir",
    [
        {
            "dir": "a",
            "manifest": "{'installable': '?'}",
        }
    ],
    indirect=True,
)
def test_manifest_type_error(addon_dir: Path) -> None:
    with pytest.raises(AddonNotFoundInvalidManifest):
        Addon.from_addon_dir(addon_dir)


def test_no_manifest(addon_dir: Path) -> None:
    (addon_dir / "__manifest__.py").unlink()
    with pytest.raises(AddonNotFoundNoManifest):
        Addon.from_addon_dir(addon_dir)


def test_no_init(addon_dir: Path) -> None:
    (addon_dir / "__init__.py").unlink()
    with pytest.raises(AddonNotFoundNoInit):
        Addon.from_addon_dir(addon_dir)


def test_addon_constructor() -> None:
    manifest = Manifest.from_dict({"name": "the addon"})
    addon = Addon(manifest, manifest_path=Path("/tmp/theaddon/__manifest__.py"))
    assert addon.name == "theaddon"
    addon = Addon(
        manifest, manifest_path=Path("/tmp/tmp/__manifest__.py"), name="theaddon"
    )
    assert addon.name == "theaddon"


@pytest.mark.parametrize(
    "addon_dir, expected",
    [
        (
            {
                "dir": "a",
                "manifest": "{}",
            },
            True,
        ),
        (
            {
                "dir": "a",
                "manifest": "{'installable': '?'}",
            },
            False,
        ),
    ],
    indirect=["addon_dir"],
)
def test_is_addon_dir_ok(addon_dir: Path, expected: bool) -> None:
    assert is_addon_dir(addon_dir) is expected



from manifestoo_core.core_addons import (
    get_core_addon_license,
    is_core_addon,
    is_core_ce_addon,
    is_core_ee_addon,
)
from manifestoo_core.odoo_series import OdooSeries


def test_is_core_ce_addon() -> None:
    assert is_core_ce_addon("base", OdooSeries.v8_0)
    assert is_core_ce_addon("base", OdooSeries.v8_0)
    assert is_core_ce_addon("account", OdooSeries.v14_0)
    assert not is_core_ce_addon("account_accountant", OdooSeries.v14_0)


def test_is_core_ee_addon() -> None:
    assert not is_core_ee_addon("base", OdooSeries.v8_0)
    assert not is_core_ee_addon("base", OdooSeries.v8_0)
    assert not is_core_ee_addon("account", OdooSeries.v14_0)
    assert is_core_ee_addon("account_accountant", OdooSeries.v14_0)


def test_is_core_addon() -> None:
    assert is_core_addon("base", OdooSeries.v8_0)
    assert is_core_addon("base", OdooSeries.v8_0)
    assert is_core_addon("account", OdooSeries.v14_0)
    assert is_core_addon("account_accountant", OdooSeries.v14_0)


def test_get_core_addon_license() -> None:
    assert get_core_addon_license("base", OdooSeries.v8_0) == "AGPL-3"
    assert get_core_addon_license("base", OdooSeries.v9_0) == "LGPL-3"
    assert get_core_addon_license("account_accountant", OdooSeries.v14_0) == "OEEL-1"
